
module metaclass;

import std.string;
import std.metastrings;
static import std.traits;
import std.typetuple;
import std.stdio;
import std.variant;
import std.stdarg;

class MetaObject( TType ) {

    this( TType object ) {
        this.object = object;
    }

    alias .Class!(TType) Class;
    
    TReturnType invoke( immutable string methodName, TReturnType = void, TParameterTypeTuple... )( TParameterTypeTuple args )
    {
        static if( is( TReturnType == void ) )
            Class.invoke!methodName( object, args );
        else
            return Class.invoke!(methodName, TReturnType)( object, args );
    }
    
    TType object;
}

class Class( TType ) {
    alias MethodTuple!( TType, __traits( allMembers, TType ) ) methodTuple;
        
    static TReturnType invoke( immutable string methodName, TReturnType = void, TParameterTypeTuple... )( TType object, TParameterTypeTuple args ) {
        foreach( i, method; methodTuple ) {
            static if( method.name == methodName ) {
                static if( is( TReturnType == void ) ) {
                    method.invoke( object, args );
                    return;
                }
                else
                    return method.invoke!TReturnType( object, args );
            }
        }
        
        assert(false);
    }
  
    static TReturnType invokeStatic( immutable string methodName, TReturnType = void, TParameterTypeTuple... )( TParameterTypeTuple args ) {
        foreach( i, method; methodTuple ) {
            static if( method.name == methodName ) {
                static if( is( TReturnType == void ) ) {
                    method.invokeStatic( args );
                    return;
                }
                else
                    return method.invokeStatic!TReturnType( args );
            }
        }
        
        assert(false);
    }
}

struct Method( TType, immutable string MethodName ) 
{
    immutable string name = MethodName;

    alias MethodOverloadTuple!( __traits( getOverloads, TType, name ) ) overloadTuple;
    
    string toString() { 
        string output = "[";
        foreach( i, overload; overloadTuple ) {
            output ~= overload.toString();
            static if( i < overloadTuple.length - 1 )
                output ~= ",";
        }
        output ~= "]";
        
        return output;
    }

	/** Invokes a method */
    static TReturnType invoke( TReturnType = void, TParameterTypeTuple... )( TType object, TParameterTypeTuple args )
    {
		// TODO: I couldn't find a decent way to write __traits( ... ) only once.
		//       Alias won't work.
        static if( is( TReturnType == void ) )
            __traits( getMember, object, name )( args );
        else
            return __traits( getMember, object, name )( args );
    }

	/** Invokes a static method */
    static TReturnType invokeStatic( TReturnType = void, TParameterTypeTuple... )( TParameterTypeTuple args )
    {
        // Invoking a static method.
		// TODO: I couldn't find a way to do this using __traits( ... ).
		//       Used mixin( ... ) instead.
        immutable string invocation = "TType." ~ name ~ "( args );";

        static if( is( TReturnType == void ) )
            mixin( invocation );
        else
            mixin( "return " ~ invocation );;
    }
}

template Tuple( E... ) {
    alias E Tuple;
}

template MethodTuple( TType, TMethodTuple... ) {
    static if( TMethodTuple.length == 0 )
        alias Tuple!() MethodTuple;
    else static if( TMethodTuple.length == 1 )
        // Testing if this really is a method.
        // TODO : I couldn't find a way to do this using __traits( ... ).
        //        Used mixin( ... ) instead.
        static if( __traits( compiles, mixin("TType." ~ TMethodTuple[0]) ) ) {
            alias Tuple!( Method!( TType, TMethodTuple ) ) MethodTuple;
        }
        else
            alias Tuple!() MethodTuple;
    else
        alias Tuple!( MethodTuple!( TType, TMethodTuple[0] ), MethodTuple!( TType, TMethodTuple[1..$] ) ) MethodTuple;
}

struct MethodOverload( alias Overload, OverloadType = typeof(Overload) )
{
    static this() {
        signature = ( isStatic ? "static " : "" ) ~ ( isFinal ? "final " : "" ) ~ returnType.stringof ~ " " ~ __traits( identifier, overload ) ~ "(";
        foreach( i, parameter; parameterTypeTuple )
        {
            signature ~= parameter.stringof;
            if( i < parameterTypeTuple.length - 1 )
                signature ~= ", ";
        }
        signature ~= ")";
    }

    string toString() { return signature; }
    
    alias Overload overload;
    
    static immutable string signature;

    static immutable bool isStatic = __traits( isStaticFunction, overload );
    static immutable bool isFinal = __traits( isFinalFunction, overload );
    static immutable bool isVirtual = __traits( isVirtualFunction, overload );

    alias std.traits.ParameterTypeTuple!(OverloadType) parameterTypeTuple;
    alias std.traits.ReturnType!(OverloadType) returnType;
};

template MethodOverloadTuple( TOverloadTuple... ) {
    static if( TOverloadTuple.length == 0 )
        alias Tuple!() MethodOverloadTuple;
    else static if( TOverloadTuple.length == 1 )
        alias Tuple!( MethodOverload!( TOverloadTuple )() ) MethodOverloadTuple;
    else
        alias Tuple!( MethodOverloadTuple!( TOverloadTuple[0] ), MethodOverloadTuple!( TOverloadTuple[1..$] ) ) MethodOverloadTuple;
}
